package ru.landgrafhomyak.bgtu.networks0.icmp

import kotlin.coroutines.Continuation
import kotlin.coroutines.resume
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.Job
import kotlinx.coroutines.TimeoutCancellationException
import kotlinx.coroutines.launch
import kotlinx.coroutines.suspendCancellableCoroutine
import kotlinx.coroutines.sync.withLock
import kotlinx.coroutines.withTimeout
import kotlinx.datetime.Clock
import kotlinx.datetime.Instant
import kotlinx.coroutines.sync.Mutex as CMutex
import ru.landgrafhomyak.bgtu.networks0.low_level.sockets.IcmpSocket
import ru.landgrafhomyak.bgtu.networks0.utilities.safeAutoClose1
import ru.landgrafhomyak.bgtu.networks0.utilities.safeAutoClose2
import ru.landgrafhomyak.utility.IntSerializers


@OptIn(ExperimentalUnsignedTypes::class)
class PingSocket : AutoCloseable {
	private val _rawSocket: IcmpSocket
	private val _mapSync: CMutex
	private val _resumeMap: HashMap<ShortKeyPair, Continuation<Unit>>
	private val _recvCoro: Job

	internal constructor(rawSocket: IcmpSocket) {
		safeAutoClose2(onAbort = rawSocket::close) {
			this._rawSocket = rawSocket
			this._resumeMap = HashMap()
			this._mapSync = CMutex()
			this._recvCoro = CoroutineScope(Dispatchers.Default).launch { this@PingSocket._recvRoutine() }
		}
	}

	private class ShortKeyPair(val identifier: UShort, val seqNo: UShort) {
		override fun equals(other: Any?): Boolean {
			if (other !is ShortKeyPair) return false
			return this.identifier == other.identifier && this.seqNo == other.seqNo
		}

		override fun hashCode(): Int {
			return ((this.identifier.toUInt() shl 16) or (this.seqNo).toUInt()).toInt()
		}

		override fun toString(): String = "<icmp ping key pair identifier=${this.identifier} seq_no=${this.seqNo}>"
	}

	private fun _generateKeyPair(): ShortKeyPair {
		val timestamp = Clock.System.now()
		return ShortKeyPair(
			identifier = (timestamp.epochSeconds ushr 13).toULong().toUShort(),
			seqNo = ((timestamp.epochSeconds shl 3) or ((((timestamp.nanosecondsOfSecond ushr 27) and 7).toLong())) and 0xFFFF).toULong().toUShort(),
		)
	}

	suspend fun ping(timeoutMillis: UInt): UInt? {
		var isLocked = true
		val start: Instant
		this._mapSync.lock()
		safeAutoClose1(finally = { if (isLocked) this._mapSync.unlock() }) {
			val key = this._generateKeyPair()
			val header = IcmpPingHeader(identifier_be = key.identifier, seqNo_be = key.seqNo)
			val serialized = header.serialize(PingSocket.randomData)
			start = Clock.System.now()
			this._rawSocket.sendRaw(serialized)
			try {
				withTimeout(timeoutMillis.toULong().toLong()) {
					suspendCancellableCoroutine<Unit> { continuation ->
						if (key in this@PingSocket._resumeMap)
							throw RuntimeException("Key duplication (shouldn't happen)")
						this@PingSocket._resumeMap[key] = continuation
						this@PingSocket._mapSync.unlock()
						isLocked = false
					}
				}
			} catch (_: TimeoutCancellationException) {
				return null
			}
		}

		val td = Clock.System.now() - start
		return td.inWholeMilliseconds.toUInt()
	}

	private suspend fun _recvRoutine() {
		try {
			// todo handle errors
			while (true) {
				val raw = this._rawSocket.recvRaw()
				if (raw.size < 8) continue
				val key = ShortKeyPair(identifier = IntSerializers.decode16beHu(raw, 4), seqNo = IntSerializers.decode16beHu(raw, 6))
				val continuation = this._mapSync.withLock { this._resumeMap.remove(key) }
				if (continuation == null) continue
				continuation.resume(Unit)
			}
		} catch (e: Throwable) {
			e.printStackTrace()
		}
	}

	override fun close() {
		TODO("Not yet implemented")
	}

	companion object {
		@Suppress("SpellCheckingInspection")
		private val randomData = "abcdefghijklmnopqrstuvwabcdefghi".encodeToByteArray().asUByteArray()
	}
}