package ru.landgrafhomyak.bgtu.networks0.low_level.multithreading

import kotlinx.cinterop.CPointer
import kotlinx.cinterop.ExperimentalForeignApi
import kotlinx.cinterop.alloc
import kotlinx.cinterop.free
import kotlinx.cinterop.nativeHeap
import kotlinx.cinterop.ptr
import platform.posix.pthread_cond_t
import platform.posix.pthread_cond_init
import platform.posix.pthread_cond_destroy
import platform.posix.pthread_cond_wait
import platform.posix.pthread_cond_broadcast
import platform.posix.pthread_cond_signal
import ru.landgrafhomyak.bgtu.networks0.low_level.c_interop_utilities.PosixUtilities
import ru.landgrafhomyak.bgtu.networks0.utilities.CloseableRefCounter
import ru.landgrafhomyak.bgtu.networks0.utilities.safeAutoClose1
import ru.landgrafhomyak.bgtu.networks0.utilities.safeAutoClose2


@OptIn(ExperimentalForeignApi::class)
actual class Condition : AutoCloseable {
	private val _refcnt: CloseableRefCounter
	val _descriptor: CPointer<pthread_cond_t>

	actual constructor() {
		this._refcnt = CloseableRefCounter("Pthreads condition was destroyed")
		this._descriptor = nativeHeap.alloc<pthread_cond_t>().ptr
		safeAutoClose2(onAbort = { nativeHeap.free(this._descriptor) }) {
			var err = pthread_cond_init(this._descriptor, null)
			if (err != 0)
				PosixUtilities.throwErrno { d -> RuntimeException("Failed to initialize pthreads condition: $d") }
		}
	}

	@OptIn(Mutex.RefcntAccess::class)
	actual fun await(mutex: Mutex) = this._refcnt.withRef {
		mutex._refcnt.withRef {
			var err = pthread_cond_wait(this._descriptor, mutex._descriptor)
			if (err != 0)
				PosixUtilities.throwErrno(err) { d -> RuntimeException("Failed to wait on pthreads condition: $d") }
		}
	}

	actual fun signalAll() = this._refcnt.withRef {
		var err = pthread_cond_broadcast(this._descriptor)
		if (err != 0)
			PosixUtilities.throwErrno(err) { d -> RuntimeException("Failed to signal pthreads condition: $d") }
	}

	actual fun signalOne() = this._refcnt.withRef {
		var err = pthread_cond_signal(this._descriptor)
		if (err != 0)
			PosixUtilities.throwErrno(err) { d -> RuntimeException("Failed to signal pthreads condition: $d") }
	}

	override fun close() {
		this._refcnt.close("There are waiting threads on this pthreads condition")
		safeAutoClose1(
			action = {
				var err = pthread_cond_destroy(this._descriptor)
				if (err != 0)
					PosixUtilities.throwErrno(err) { d -> RuntimeException("Failed to destroy pthreads condition: $d") }
			},
			finally = { nativeHeap.free(this._descriptor) }
		)
	}
}