package ru.landgrafhomyak.bgtu.networks0.low_level.multithreading

import kotlinx.atomicfu.AtomicRef
import kotlinx.atomicfu.atomic
import kotlinx.atomicfu.update
import kotlinx.cinterop.CPointer
import kotlinx.cinterop.ExperimentalForeignApi
import kotlinx.cinterop.StableRef
import kotlinx.cinterop.asStableRef
import kotlinx.cinterop.staticCFunction
import ru.landgrafhomyak.bgtu.networks0.low_level.c_interop_utilities.PosixUtilities
import ru.landgrafhomyak.bgtu.networks0.utilities.safeAutoClose1
import ru.landgrafhomyak.bgtu.networks0.utilities.safeAutoClose2
import ru.landgrafhomyak.bgtu.networks0.utilities.safeAutoClose2e
import ru.landgrafhomyak.bgtu.networks0.utilities.compareAndExchange

@OptIn(ExperimentalForeignApi::class)
actual class Thread : AutoCloseable {
	actual fun interface Routine {
		actual fun run()
	}

	internal enum class State {
		PENDING, STARTING, RUNNING, FINISHED, CLOSED
	}


	internal inner class ThreadBootstrapContext(val routine: Routine) {
		val startSignal = CountDownLatch(1uL)
		val threadState by this@Thread::_state
		var exitedWithError: Throwable? = null
	}

	private val _threadBindings: _PthreadsThreadBindings
	private val _bootstrapArgRef: StableRef<ThreadBootstrapContext>
	private val _state: AtomicRef<State>

	actual constructor(routine: Routine) {
		this._state = atomic(State.PENDING)
		this._bootstrapArgRef = StableRef.create(ThreadBootstrapContext(routine))
		safeAutoClose2(onAbort = this._bootstrapArgRef::dispose) {
			this._threadBindings = _PthreadsThreadBindings()
			safeAutoClose2(onAbort = this._threadBindings::free) {
				var err = this._threadBindings.create(
					null,
					staticCFunction({ arg -> return@staticCFunction Thread._threadBootstrap(arg) }),
					this._bootstrapArgRef.asCPointer()
				)
				if (err != 0)
					PosixUtilities.throwErrno(err) { d -> RuntimeException("Failed to create pthreads thread: $d") }
			}
		}

	}

	actual fun start() {
		when (this._state.compareAndExchange(State.PENDING, State.STARTING)) {
			State.CLOSED -> throw IllegalStateException("Pthreads thread is destroyed")
			State.FINISHED -> throw IllegalStateException("Pthreads thread was started and already finished")
			State.STARTING, State.RUNNING -> throw IllegalStateException("Pthreads thread already running")
			State.PENDING -> {}
		}
		safeAutoClose2(onAbort = { this._state.value = State.PENDING }) {
			this._bootstrapArgRef.get().startSignal.decrement()
		}
	}

	actual fun join(): Throwable? {
		when (this._state.value) {
			State.CLOSED -> throw IllegalStateException("Pthreads thread is destroyed")
			State.FINISHED -> return this._bootstrapArgRef.get().exitedWithError
			State.STARTING, State.RUNNING, State.PENDING -> {}
		}
		var err = this._threadBindings.join(null)
		if (err != 0)
			PosixUtilities.throwErrno(err) { d -> RuntimeException("Failed to create pthreads thread") }
		return this._bootstrapArgRef.get().exitedWithError
	}

	override fun close() {
		this._state.update { o ->
			when (this._state.value) {
				State.CLOSED -> throw IllegalStateException("Pthreads thread is destroyed")
				State.STARTING, State.RUNNING -> throw IllegalStateException("Can't destroy pthreads thread while it running")
				State.FINISHED, State.PENDING -> {}
			}
			return@update State.CLOSED
		}

		safeAutoClose1(
			action = {
				var err = this._threadBindings.join(null)
				if (err != 0)
					PosixUtilities.throwErrno(err) { d -> RuntimeException("Failed to destroy pthreads thread: $d") }
			},
			finally = {
				safeAutoClose1(
					action = { this._threadBindings.free() },
					finally = { this._bootstrapArgRef.dispose() }
				)
			}
		)
	}

	companion object {
		private fun _threadBootstrap(arg: CPointer<*>?): CPointer<*>? {
			if (arg == null) return null
			val context: ThreadBootstrapContext
			try {
				context = arg.asStableRef<ThreadBootstrapContext>().get()
			} catch (_: Throwable) {
				return null
			}
			try {
				safeAutoClose2e(
					onAbort = { e ->
						context.exitedWithError = e
						context.threadState.value = State.FINISHED
					},
					onSuccess = { context.threadState.value = State.FINISHED },
					action = {
						context.startSignal.await()
						context.routine.run()
					}
				)
			} catch (_: Throwable) {
			}
			return null
		}
	}
}