package ru.landgrafhomyak.bgtu.networks0.low_level.sockets

import kotlinx.cinterop.CValue
import kotlinx.cinterop.ExperimentalForeignApi
import kotlinx.cinterop.addressOf
import kotlinx.cinterop.memScoped
import kotlinx.cinterop.usePinned
import kotlinx.coroutines.sync.Mutex as CMutex
import kotlinx.coroutines.sync.withLock
import platform.posix.AF_INET
import platform.posix.AF_INET6
import platform.posix.EAGAIN
import platform.posix.EMSGSIZE
import platform.posix.EPERM
import platform.posix.EWOULDBLOCK
import platform.posix.F_SETFL
import platform.posix.IPPROTO_ICMP
import platform.posix.IPPROTO_ICMPV6
import platform.posix.MSG_NOSIGNAL
import platform.posix.MSG_PEEK
import platform.posix.MSG_TRUNC
import platform.posix.O_NONBLOCK
import platform.posix.SOCK_RAW
import platform.posix.connect
import platform.posix.errno
import platform.posix.fcntl
import platform.posix.sockaddr
import platform.posix.sockaddr_in
import platform.posix.sockaddr_in6
import platform.posix.socket
import platform.posix.ssize_t
import platform.posix.send as send_lowlevel
import platform.posix.recv as recv_lowlevel
import platform.posix.close as close_lowlevel
import ru.landgrafhomyak.bgtu.networks0.low_level.c_interop_utilities.CUtilities
import ru.landgrafhomyak.bgtu.networks0.low_level.c_interop_utilities.PosixUtilities
import ru.landgrafhomyak.bgtu.networks0.utilities.CloseableRefCounter
import ru.landgrafhomyak.bgtu.networks0.utilities.safeAutoClose2

@Suppress("CanBeVal")
@OptIn(ExperimentalForeignApi::class)
internal abstract class NonblockingIcmpSocketImpl : IcmpSocket {
	protected val _socketFd: Int
	private val __rSync: CMutex
	private val __wSync: CMutex
	private val __refcnt: CloseableRefCounter

	protected constructor(ipv4: CValue<sockaddr_in>) {
		this.__rSync = CMutex()
		this.__wSync = CMutex()
		this.__refcnt = CloseableRefCounter("ICMP socket was closed")
		this._socketFd = this.__createSocket(AF_INET, IPPROTO_ICMP, ipv4, CUtilities.sizeOfUI<sockaddr_in>(), "IPv4")
	}

	protected constructor(ipv6: CValue<sockaddr_in6>) {
		this.__rSync = CMutex()
		this.__wSync = CMutex()
		this.__refcnt = CloseableRefCounter("ICMP socket was closed")
		this._socketFd = this.__createSocket(AF_INET6, IPPROTO_ICMPV6, ipv6, CUtilities.sizeOfUI<sockaddr_in6>(), "IPv6")
	}

	@Suppress("FunctionName")
	private fun __createSocket(addrFamily: Int, protocol: Int, addrValue: CValue<*>, addrSize: UInt, protoName: String): Int {
		val sock = socket(addrFamily, SOCK_RAW, protocol)
		if (sock < 0) {
			if (errno == EPERM)
				PosixUtilities.throwErrno { d -> RuntimeException("Failed to create ICMP socket over ${protoName}: $d (maybe you need to run this program as superuser)") }
			else
				PosixUtilities.throwErrno { d -> RuntimeException("Failed to create ICMP socket over ${protoName}: $d") }
		}
		safeAutoClose2(onAbort = { if (0 != close_lowlevel(this._socketFd)) PosixUtilities.throwErrno { d -> RuntimeException("Failed to close ICMP socket: $d") } }) {
			var err = memScoped {
				@Suppress("UNCHECKED_CAST")
				return@memScoped connect(sock, addrValue as CValue<sockaddr>, addrSize)
			}
			if (err != 0)
				PosixUtilities.throwErrno { d -> RuntimeException("Failed to bind ICMP socket to provided $protoName address: $d") }

			err = fcntl(sock, F_SETFL, O_NONBLOCK)
			if (err != 0)
				PosixUtilities.throwErrno { d -> RuntimeException("Failed to switch ICMP socket to non-blocking mode: $d") }
		}
		return sock

	}

	@Suppress("FunctionName")
	protected abstract suspend fun _waitForWrite()

	@Suppress("FunctionName")
	protected abstract suspend fun _waitForRead()

	override suspend fun sendRaw(data: UByteArray) {
		this.__refcnt.withRef {
			this.__wSync.withLock {
				polling@ while (true) {
					this._waitForWrite()
					val sentCount = data.usePinned { pinnedData ->
						return@usePinned send_lowlevel(this._socketFd, pinnedData.addressOf(0), data.size.toULong(), MSG_NOSIGNAL)
					}
					val err = errno
					if (sentCount < 0) {
						when (err) {
							EAGAIN, EWOULDBLOCK -> continue@polling
							EMSGSIZE -> throw IllegalArgumentException("Data size bigger than can be sent by one packet")
							else -> PosixUtilities.throwErrno(err) { d -> RuntimeException("Failed to send message ove ICMP socket: $d") }
						}
					}
					return
				}
			}
		}
	}

	override suspend fun recvRaw(): UByteArray {
		this.__refcnt.withRef {
			this.__rSync.withLock {
				polling@ while (true) {
					this._waitForRead()
					val bytesAvailable: ssize_t = recv_lowlevel(
						this._socketFd, null, 0u,
						MSG_NOSIGNAL or MSG_PEEK or MSG_TRUNC
					)

					if (bytesAvailable == 0L) continue@polling
					if (bytesAvailable < 0) {
						when (errno) {
							EAGAIN, EWOULDBLOCK -> continue@polling
							else -> PosixUtilities.throwErrno { d -> RuntimeException("Failed to send message over ICMP socket: $d") }
						}
					}

					val data = UByteArray(bytesAvailable.toInt())
					var receivedCount = data.usePinned { pinnedData ->
						return@usePinned recv_lowlevel(this._socketFd, pinnedData.addressOf(0), data.size.toULong(), MSG_NOSIGNAL)
					}
					if (receivedCount < 0) {
						when (errno) {
							EAGAIN, EWOULDBLOCK -> continue@polling
							else -> PosixUtilities.throwErrno { d -> RuntimeException("Failed to send message over ICMP socket: $d") }
						}
					}
					return data.sliceArray(((data[0] and 0xFu).toUInt().toInt() + 15)..<data.size)
				}
				@Suppress("KotlinUnreachableCode")
				throw RuntimeException("unreachable")
			}
		}
	}

	protected open fun _free() {
		var err = close_lowlevel(this._socketFd)
		if (err != 0)
			PosixUtilities.throwErrno { d -> RuntimeException("Failed to close ICMP socket: $d") }
	}

	override fun close() {
		this.__refcnt.close("This socket still in use")
		this._free()
	}
}