@file:OptIn(ExperimentalContracts::class)

package ru.landgrafhomyak.bgtu.networks0.utilities

import kotlin.contracts.ExperimentalContracts
import kotlin.contracts.InvocationKind
import kotlin.contracts.contract
import kotlinx.atomicfu.AtomicLong
import kotlinx.atomicfu.atomic
import kotlinx.atomicfu.update

class CloseableRefCounter(private val _errMessage: String) {
	private val _value: AtomicLong = atomic(0L)

	fun throwClosed() {
		throw IllegalStateException(this._errMessage)
	}

	fun incref() {
		this._value.update { o ->
			if (o < 0) this.throwClosed()
			return@update o + 1
		}
	}

	inline fun <R> tryIncref(block: () -> R): R {
		contract {
			callsInPlace(block, InvocationKind.EXACTLY_ONCE)
		}
		this.incref()
		return safeAutoClose2(onAbort = this::decref, action = block)
	}

	fun checkNotClosed() {
		if (this._value.value < 0) this.throwClosed()
	}

	fun decref() {
		this._value.update(Long::dec)
	}

	inline fun <R> tryDecref(block: () -> R): R {
		contract {
			callsInPlace(block, InvocationKind.EXACTLY_ONCE)
		}
		this.checkNotClosed()
		return safeAutoClose2(onSuccess = this::decref, action = block)
	}

	fun close(errExistRefs: String) {
		val state = this._value.compareAndExchange(0, -1)
		when {
			state > 0 -> throw IllegalStateException(errExistRefs)
			state < 0 -> this.throwClosed()
		}
	}

	inline fun <R> withRef(protected: () -> R): R {
		this.incref()
		return safeAutoClose1(finally = this::decref, action = protected)
	}

	override fun toString(): String {
		val refcntCached = this._value.value
		if (refcntCached < 0)
			return "<ref counter [closed]>"
		else
			return "<ref counter [${refcntCached}]>"
	}
}